from ever_playground import Cell as C
from ever_playground import Builder as B
from ever_playground import Slice as S
from ever_playground import Dictionary as D
from ever_playground import parse_smc_addr

def expect(expected, v):
    if not expected == v:
        raise Exception("{} != {}".format(expected, v))

b = B()
b.i(8, 0x12).i(8, 0x34)
s = S(b.finalize())
expect(0x1234, s.u(16))

val1 = -123
val2 = 0xdeadbeefcafebabe
s = B().i(32, val1).i(64, val2).slice()
expect(val1, s.i(32))
expect(False, s.is_empty())
expect(val2, s.u(64))
expect(True, s.is_empty())

val3 = 0xffffffff
s = S(B().i(32, val3).finalize())
expect(-1, s.i(23))
expect(9, s.remaining_bits())

d1_wc, d1_addr = parse_smc_addr("0:000169b042c37962027e58de0dbaa0b85f5d032f37d8333e3cdfdcc7918ae00a")
d2_wc, d2_addr = parse_smc_addr("0:bc43df2056abee4c1a443fbfcfede0ba90d214c77322167fc08ce48920c17c1b")
d3_wc, d3_addr = parse_smc_addr("0:d4a50c1a849a4742214e751977af268269e6eeae064ce800634acea241d430d3")

s1 = B().i(32, d1_wc).i(256, d1_addr).slice()
s2 = B().i(32, d2_wc).i(256, d2_addr).slice()
# s3 = B().i(32, d3_wc).i(256, d3_addr).slice()
b3 = B()
b3.i(32, d3_wc)
b3.i(256, d3_addr)
s3 = b3.slice()

dict = D(288)
dict.add_kv_slice(288, s1)
dict = dict.add_kv_slice(288, s2)
dict.add_kv_slice(288, s3)
#dict = D(288).add_kv_slice(288, s1).add_kv_slice(288, s2).add_kv_slice(288, s3)

expect(S(B().finalize()), dict.get(s1))
expect(None, dict.get(B().i(288, 0).slice()))

dict_cell1 = dict.serialize().finalize()
#print(dict_cell1)

dict = D.deserialize(288, S(dict_cell1))

empty = C("")
expect(empty, B().finalize())
expect(0x96a296d224f285c67bee93c30f8a309157f0daa35dc5b87e410b78630a09cfc7, empty.repr_hash())

c = \
C("c_",
    C("c20",
        C("bfc000b4d82161bcb1013f2c6f06dd505c2fae81979bec199f1e6fee63c8c570054_"),
        C("2_",
            C("bfbc43df2056abee4c1a443fbfcfede0ba90d214c77322167fc08ce48920c17c1b"),
            C("bf94a50c1a849a4742214e751977af268269e6eeae064ce800634acea241d430d3"))))
expect(c, dict_cell1)

dict_bytes = dict_cell1.write(0)
dict_cell2 = C.read(dict_bytes)
#print(dict_cell2)

expect(dict_cell1, dict_cell2)

#open("test.boc", "wb").write(dict_cell.write())
#cell = C.read(open("test.boc", "rb").read())



from ever_playground import *

def a(text: str) -> Slice:
    return Slice(assemble(text))

def make_while(cc: Continuation, c0: Continuation, body_text: str, cond_text: str) -> Continuation:
    body = a(body_text)
    cond = Continuation(
        typ = ContinuationType.create_ordinary(),
        code = a(cond_text)
    )
    while_ = Continuation(typ = ContinuationType.create_while(body, cond.code))
    cc.savelist.put(0, c0)
    while_.savelist.put(0, cc)
    cond.savelist.put(0, while_)
    return cond

def runcont(entry: Continuation, gas_limit: int = 1_000_000, gas_credit: int = 10_000, trace = False) -> VmResult:
    regs = SaveList()
    regs.put(0, entry)
    gas = Gas(gas_limit, gas_credit)
    return runvm_generic(VmState(Continuation(), regs, gas), trace = trace)

quit0 = Continuation(typ = ContinuationType.create_quit(500))
rest = Continuation(typ = ContinuationType.create_ordinary(), code = a("RET"))

program = make_while(rest, quit0, "INC", "DUP PUSHINT 5 LESS")
program.stack.append(0)

r = runcont(program)
assert(r.exit_code == 500)
assert(r.state.cc.stack == [5])



import sys
import time
import argparse
from ever_playground import Builder, Cell, Currency, Dictionary, Slice, \
    parse_smc_addr, load_address, load_keypair, ed25519_sign

def abort(message: str):
    raise Exception(message)

mode_default = 3
timeout_default = 60

argparser = argparse.ArgumentParser(prog = sys.argv[0], description =
    "Creates a request with up to 254 orders loaded from <order-file> to high-load (sub)wallet created by new-highload-v2-wallet.fif, with private key loaded from file <filename-base>.pk "
    "and address from <filename-base><subwallet-id>.addr, and saves it into <savefile>.boc ('wallet-query.boc' by default) "
    "<order-file> is a text file with lines `SEND <dest-addr> <amount>`")
argparser.add_argument("filename_base", help = "prefix for input file names")
argparser.add_argument("subwallet_id", type = int, help = "subwallet id")
argparser.add_argument("order_file", help = "name of file with a list of orders")
argparser.add_argument("--savefile", default = "wallet-query", help = "output file name w/o extension")
argparser.add_argument("-n", "--no-bounce", action = "store_false", help = "clears bounce flag")
argparser.add_argument("-b", "--force-bounce", action = "store_true", help = "forces bounce flag")
argparser.add_argument("-t", "--timeout", type = int, action = "store", default = timeout_default,
    help = f"sets expiration timeout in seconds ({timeout_default} by default)")
argparser.add_argument("-m", "--mode", type = int, action = "store", default = mode_default,
    help = f"sets transfer mode (0..255) for SENDRAWMSG ({mode_default} by default)")
args = argparser.parse_args()

file_base: str = args.filename_base
subwallet_id: int = args.subwallet_id
if subwallet_id.bit_length() > 32:
    argparser.print_help()
    exit(1)
order_file: str = args.order_file
savefile: str = args.savefile
allow_bounce: bool = args.no_bounce
force_bounce: bool = args.force_bounce
timeout: int = args.timeout
send_mode: int = args.mode

wallet_wc, wallet_addr = load_address(file_base + str(subwallet_id) + ".addr")
print(f"Source wallet address = {wallet_wc}:{wallet_addr:064x}")

_, wallet_pk = load_keypair(file_base + ".pk")

orders = Dictionary(16)
order_number = 0

def add_order(order: Slice):
    global orders, order_number
    if order_number > 254:
        abort("more that 254 orders")
    key = Slice(Builder().i(16, order_number).finalize())
    if orders.get(key) != None:
        abort("cannot add order to dictionary")
    orders.add(key, order)
    order_number += 1

def append_msg_body(b: Builder, body: Cell):
    body_slice = Slice(body)
    if b.fits(body_slice, 1, 0):
        b.i(1, 0).s(body_slice)
    else:
        b.i(1, 1).r(body)

def create_int_msg(body: Cell, bounce: bool, wc: int, addr: int, ng: Currency) -> Cell:
    b = Builder().ib("01").i(1, bounce).ib("000100").i(8, wc).i(256, addr) \
        .b(ng.serialize()).i(9 + 64 + 32 + 1, 0)
    append_msg_body(b, body)
    return b.finalize()

def create_simple_transfer(address: str, ng: Currency, bounce: bool) -> Cell:
    wc, addr = parse_smc_addr(address)
    bounce = bounce or force_bounce and allow_bounce
    print(f"Transferring {ng} to account {wc}:{addr:064x} bounce={bounce}")
    return create_int_msg(Builder().i(32, 0).finalize(), bounce, wc, addr, ng)

def create_order(c: Cell, send_mode: int) -> Cell:
    return Builder().i(8, send_mode).r(c).finalize()

def send(address: str, ng: Currency, bounce: bool):
    t = create_simple_transfer(address, ng, bounce)
    order = create_order(t, send_mode)
    add_order(Slice(order))

with open(order_file, "r") as file:
    for line in file:
        chunks = line.split(" ")
        assert(len(chunks) == 4)
        assert(chunks[0] == "SEND")
        address = chunks[1]
        bounce = bool(chunks[2])
        ng = Currency.from_str(chunks[3])
        send(address, ng, bounce)

now = 1684770872 # int(time.time())
orders_cell = orders.serialize().finalize()
query_id = ((now + timeout) << 32) + (orders_cell.repr_hash() & 0xffffffff)

signing_message = Builder().i(32, subwallet_id).i(64, query_id).s(Slice(orders_cell)).finalize()
print(f"signing message: {signing_message}")

signature = ed25519_sign(signing_message.repr_hash().to_bytes(32, "big"), wallet_pk)
ext = Builder().ib("1000100").i(8, wallet_wc).i(256, wallet_addr).b(Currency(0).serialize()) \
    .ib("00").y(signature).s(Slice(signing_message)).finalize()
print(f"resulting external message: {ext}")
ext_bytes = ext.write(2)
print(f"{ext_bytes.hex()}")
print(f"query_id is {query_id} = 0x{query_id:x}")
open(savefile + ".boc", "wb").write(ext_bytes)
print(f"(Saved to file {savefile}.boc)")



from ever_playground import *

code0 = assemble("""
    ; stack to be preserved, here 1 slot
    PUSHINT 111
    ; try cont
    PUSHCONT {
        INC ; update stack [111]
        PUSHINT 222
        ; compiler knows added stack, here it is [222]
        ; drop it
        DROP
        ; pass remaining (and updated) stack to c2 (catch cont)
        PUSHCTR c2
        DEPTH DEC ; r <= 255
        PUSHINT 0 ; n
        SETCONTVARARGS
        POPCTR c2
        ; fire an exception
        THROW 123
    }
    ; catch cont
    PUSHCONT {
        ; do something with exception pair; here, just drop
        DROP2
        ; do somethig else
        PUSHINT 333
    }
    TRY
    PUSHINT 444
""")

# int v = 111;
# try {
#   inc(v);
#   push(222);
#   throw(123);
# } catch (val, exccode) {
#   push(333);
# }
# push(444);
code = assemble("""
    ; stack to be preserved; here, it's 1 slot
    PUSHINT 111
    DEPTH
    ; try cont
    PUSHCONT {
        ; user code 1
        INC
        PUSHINT 222
        PUSHINT 333
        ; pass entire stack to c2 (catch cont) + current stack depth
        DEPTH
        PUSHCTR c2
        DEPTH DEC ; r < 256
        PUSHINT -1 ; nargs
        SETCONTVARARGS
        POPCTR c2
        ; user code 2: fire an exception
        THROW 123
    }
    ; catch cont
    PUSHCONT {
        ; leave only (1) stack being preserved and (2) exc pair
        ; here goes the magic of [n, x1..xn, y1..yk, k+n, v, e] -> [x1..xn, v, e]
        DEPTH DEC ROLLX
        ROLL 3 SUBR DUP
        PUSHINT 2 ADD
        PUSHINT 1
        REVX
        DROPX
        SWAP
        ; do smth with exception pair; here, just drop it
        DROP2
        PUSHINT 444
    }
    XCHG s1, s2
    SETCONTARGS 1, -1
    TRY
    PUSHINT 555
""")

code = assemble("""
    PUSHINT 111
    PUSHCONT {
        PUSHINT 222
        PUSHINT 333
        PUSHCTR c2
        SETCONTARGS 3, -1
        POPCTR c2
        THROW 123
    }
    PUSHCONT {
        PUSHINT 444
    }
    TRY
""")

code = assemble("""
    PUSHCONT {
        PUSHCONT {
            TRUE
        }
        PUSHCONT {
            SAVEALT c2 ;; c1.savelist(2) = c2
            PUSHCONT {
                THROW 111 ;RETALT
            }
            PUSHCONT {
                RETALT
            }
            .blob xf2fe ;; TRYKEEP
        }
        WHILEBRK
    }
    CALLX
    THROW 228
""")

r = runvm(Slice(code), [], capabilities = 0x0020_0000, trace = True)
assert(r.exit_code == 228)
exit(0)

# d c1 c2 - c1 d c2 ; a b c - b a c
# XCHG s1, s2

# [k, x1..xn, y1..yk, v, e] -> [x1..xn, v, e]

r = runvm(Slice(code), [], capabilities = 0x0020_0000, trace = True)
print(r.stack)
assert(r.stack == [112, 444, 555])

code = assemble("""
    PUSHCTR c0
    PUSHCONT {
        JMPX
    }
    PUSHCONT {
    }
    TRY
    THROW 500
""")

code = assemble("""
    PUSHCONT {
        PUSHREF {
            .blob x800000000fffffc00000000002_
            .cell {
            }
        }
        CTOS LDCONT ENDS
        JMPX
    }
    PUSHCONT {
    }
    TRY
    THROW 500
""")



#!/usr/bin/env python3

import filecmp
import subprocess
import os

from ever_playground import Cell

ton_src = "/tonlabs/ton"
ton_build = ton_src + "/build"
fift = [ton_build + "/crypto/fift", "-I", ton_src + "/crypto/fift/lib", "-s"]

def run(params):
    p = subprocess.run(params)
    assert(0 == p.returncode)

def remove(filename: str):
    try:
        os.remove(filename)
    except OSError:
        pass

def assert_identical_files(filename1: str, filename2: str):
    res = filecmp.cmp(filename1, filename2)
    remove(filename1)
    remove(filename2)
    if not res:
        raise Exception("outputs are not identical")

def run_basics():
    run(["python3", "examples/basics.py"])

def run_runvm():
    run(["python3", "examples/runvm.py"])

def test_recover_stake():
    script = "examples/recover-stake"
    base = "recover-query"
    py_output = "py-" + base + ".boc"
    fift_output = "fift-" + base + "boc"

    run(["python3", script + ".py", py_output])
    run(fift + [script + ".fif", fift_output])

    assert_identical_files(py_output, fift_output)

def test_testgiver():
    script = "examples/testgiver"
    base = "testgiver-query"
    py_output = "py-" + base
    fift_output = "fift-" + base
    common = ["13:7777777777777777777777777777777777777777777777777777777777777777", "1234", "15.7"]

    run(["python3", script + ".py"] + common + [py_output])
    run(fift + [script + ".fif"] + common + [fift_output])

    assert_identical_files(py_output + ".boc", fift_output + ".boc")

def test_validator_elect_req():
    script = "examples/validator-elect-req"
    base = "validator-to-sign.bin"
    py_output = "py-" + base
    fift_output = "fift-" + base
    common = ["-1:7777777777777777777777777777777777777777777777777777777777777777", "1684501715", "33.333",
        "8888888888888888888888888888888888888888888888888888888888888888"]
    
    run(["python3", script + ".py"] + common + [py_output])
    run(fift + [script + ".fif"] + common + [fift_output])

    assert_identical_files(py_output, fift_output)

def test_highload_wallet_v2():
    curdir = os.getcwd()
    os.chdir("examples/highload")

    script = "highload-wallet-v2"
    base = "wallet-query"
    py_output = "py-" + base
    fift_output = "fift-" + base

    run(["python3", script + ".py", "sample", "27172", "order-list-py", "--savefile", py_output])
    run(fift + [script + ".fif", "sample", "27172", "order-list-fift", fift_output])

    # boc serialization seems different, compare deser-ed cells instead
    # assert_identical_files(py_output + ".boc", fift_output + ".boc")
    p = Cell.read(open(py_output + ".boc", "rb").read())
    f = Cell.read(open(fift_output + ".boc", "rb").read())
    remove(py_output + ".boc")
    remove(fift_output + ".boc")
    assert(p == f)

    os.chdir(curdir)

def test_ifjmp_perf_eval():
    run(["python3", "examples/ifjmp-perf-eval.py"])

def run_runcont():
    run(["python3", "examples/runcont.py"])

run_basics()
run_runvm()
run_runcont()
test_recover_stake()
test_testgiver()
test_validator_elect_req()
test_highload_wallet_v2()
test_ifjmp_perf_eval()



# #!/usr/bin/fift -s
# "TonUtil.fif" include

import sys
import time

from ever_playground import Builder as B

# { ."usage: " @' $0 type ." [<savefile>]" cr
#   ."Creates the message body to be sent from a validator controlling smart contract (wallet) to recover its share of unfrozen stakes and bonuses." cr
#   ."The result is saved into <savefile> (`recover-query.boc` by default) and output in hexadecimal form, to be sent later as the body of a message from the wallet to elections smart contract, along with a small value (say, one Gram) to cover forwarding and processing fees" cr 1 halt
# } : usage

def usage():
    print("Usage: {} [<savefile>]".format(sys.argv[0]))
    print("Creates the message body to be sent from a validator controlling smart contract (wallet) to recover its share of unfrozen stakes and bonuses.")
    print("The result is saved into <savefile> (`recover-query.boc` by default) and output in hexadecimal form, to be sent later as the body of a message from the wallet to elections smart contract, along with a small value (say, one Gram) to cover forwarding and processing fees")
    exit(1)

# $# dup 0 < swap 1 > or ' usage if
argc = len(sys.argv) - 1
if argc > 1:
    usage()

# def? $1 { @' $1 } { "recover-query.boc" } cond constant output_fname
output_filename = "recover-query.boc"
if argc > 0:
    output_filename = sys.argv[1]

# now constant query_id
query_id = 1684770872 # int(time.time())

# ."query_id for stake recovery message is set to " query_id . cr
print(f"query_id for stake recovery message is set to {query_id}")

# <b x{47657424} s, query_id 64 u, b>
# cr ."Message body is " dup <s csr. cr
body = B().x("47657424").i(64, query_id).finalize()
print(f"Message body is {body}")

# 2 boc+>B output_fname tuck B>file ."Saved to file " type cr
open(output_filename, "wb").write(body.write(2))
print(f"Saved to file {output_filename}")



# #!/usr/bin/fift -s
# "TonUtil.fif" include

import sys
from ever_playground import Builder as B, Slice as S, Currency, parse_smc_addr, parse_load_address

# { ."usage: " @' $0 type ." <dest-addr> <seqno> <amount> [<savefile>]" cr
#   ."Creates a request to TestGiver and saves it into <savefile>.boc" cr
#   ."('testgiver-query.boc' by default)" cr 1 halt
# } : usage
def usage():
    print(f"usage: {sys.argv[0]} <dest-addr> <seqno> <amount> [<savefile>]")
    print("Creates a request to TestGiver and saves it into <savefile>.boc")
    print("('testgiver-query.boc' by default)")
    exit(1)

# $# 3 - -2 and ' usage if
argc = len(sys.argv) - 1
if not (3 <= argc and argc <= 4):
    usage()

# // "testgiver.addr" load-address 
# Masterchain 0xfcb91a3a3816d0f7b8c2c76108b8a9bc5a6b7a55bd79f8ab101c52db29232260
# 2constant giver_addr
#  ."Test giver address = " giver_addr 2dup .addr cr 6 .Addr cr
giver_address = "-1:fcb91a3a3816d0f7b8c2c76108b8a9bc5a6b7a55bd79f8ab101c52db29232260"
giver_wc, giver_addr = parse_smc_addr(giver_address)
print(f"Test giver address = {giver_address}")

# $1 true parse-load-address =: bounce 2=: dest_addr
# $2 parse-int =: seqno
# $3 $>GR =: amount
# def? $4 { @' $4 } { "testgiver-query" } cond constant savefile
dest_address = sys.argv[1]
dest_wc, dest_addr = parse_load_address(dest_address)
bounce = True # TODO?
seqno = int(sys.argv[2])
amount = Currency.from_str(sys.argv[3])
savefile = "testgiver-query"
if argc > 3:
    savefile = sys.argv[4]

# ."Requesting " amount .GR ."to account "
# dest_addr 2dup bounce 7 + .Addr ." = " .addr
# ."seqno=0x" seqno x. ."bounce=" bounce . cr
print(f"Requesting {amount} to account {dest_address} seqno {hex(seqno)} bounce {bounce}")

# // create a message (NB: 01b00.., b = bounce)
# <b b{01} s, bounce 1 i, b{000100} s, dest_addr addr, 
#    amount Gram, 0 9 64 32 + + 1+ 1+ u, 0 32 u, "GIFT" $, b>
# <b seqno 32 u, 1 8 u, swap ref, b>
# dup ."enveloping message: " <s csr. cr
hint = bytes("GIFT", "utf8")
c1 = B().ib("01").i(1, bounce).ib("000100").i(8, dest_wc).i(256, dest_addr) \
    .b(amount.serialize()).i(9 + 64 + 32 + 1 + 1, 0).i(32, 0).y(hint) \
    .finalize()
message = B().i(32, seqno).i(8, 1).r(c1).finalize()
print(f"enveloping message: {message}")

# <b b{1000100} s, giver_addr addr, 0 Gram, b{00} s,
#    swap <s s, b>
# dup ."resulting external message: " <s csr. cr
# 2 boc+>B dup Bx. cr
# savefile +".boc" tuck B>file
# ."(Saved to file " type .")" cr
ext_message = B().ib("1000100").i(8, giver_wc).i(256, giver_addr) \
    .b(Currency(0).serialize()).ib("00").s(S(message)).finalize()
print(f"resulting external message: {ext_message}")
ext_message_bytes = ext_message.write(2) # include crc32
print(ext_message_bytes.hex())

output_filename = savefile + ".boc"
open(output_filename, "wb").write(ext_message_bytes)
print(f"(Saved to file {output_filename})")



import sys
from ever_playground import *

# capIhr = 1
capCreateStats = 2
capBounceMsgBody = 4
capReportVersion = 8
# capSplitMergeTransactions = 16
capShortDequeue = 32
# сapFastStorageStat = 128
# capInitCodeHash = 256
# capOffHypercube = 512
# capMyCode = 1024
# capFixTupleIndexBug 4096

def prepare_param8(version: int, capabilities: int) -> Cell:
    return Builder().x("c4").i(32, version).i(64, capabilities).finalize()

# create new value for config param #8
param8 = prepare_param8(20, capCreateStats or capBounceMsgBody or capReportVersion or capShortDequeue)
# check the validity of this value
if not is_valid_config(8, param8):
    raise Exception("not a valid value for chosen configuration parameter")

print(f"Serialized value = {param8}")
filename = sys.argv[1]
open(filename, "wb").write(param8.write(2))
print(f"Saved to file {filename}")

# { dup -1000 = { drop <s ref@ <s 12 u@ 0xFF0 = } {
#   dup -1001 = { drop <s ref@ <s 12 u@ 0xFF0 = } {
#   over null? { 2drop true } {
#   config-valid?
#   } cond } cond } cond
# } : is-valid-config?
